from PyQt5.QtWidgets import (
    QDialog, QWidget, QLabel, QSpinBox, 
    QComboBox, QDialogButtonBox, QGridLayout
)

from plover_next_stroke.next_stroke_config import NextStrokeConfig
from plover_next_stroke.sorting import SortingType, sorting_descriptions


class ConfigUI(QDialog):

    def __init__(self, temp_config: NextStrokeConfig, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.temp_config = temp_config
        self.setup_window()

    def setup_window(self) -> None:
        self.resize(350, 200)

        self.row_height_label = QLabel(self)
        self.row_height_label.setText("Row Height")

        self.row_height_box = QSpinBox(self)
        self.row_height_box.setValue(self.temp_config.row_height)
        self.row_height_box.setRange(10, 100)

        self.page_len_label = QLabel(self)
        self.page_len_label.setText("List Length")
        
        self.page_len_box = QSpinBox(self)
        self.page_len_box.setValue(self.temp_config.page_len)
        self.page_len_box.setRange(1, 30)

        self.sorting_type_label = QLabel(self)
        self.sorting_type_label.setText("Display Order")

        self.sorting_type_box = QComboBox(self)
        self.sorting_type_box.addItems(sorting_descriptions)
        self.sorting_type_box.setCurrentIndex(self.temp_config.sorting_type.value)

        self.button_box = QDialogButtonBox(
            (
                QDialogButtonBox.Cancel | 
                QDialogButtonBox.Ok
            ),
            parent=self
        )
        self.button_box.rejected.connect(self.reject)
        self.button_box.accepted.connect(self.save_settings)

        self.layout = QGridLayout()
        self.layout.addWidget(self.row_height_label, 0, 0)
        self.layout.addWidget(self.row_height_box, 0, 1)
        self.layout.addWidget(self.page_len_label, 1, 0)
        self.layout.addWidget(self.page_len_box, 1, 1)
        self.layout.addWidget(self.sorting_type_label, 2, 0)
        self.layout.addWidget(self.sorting_type_box, 2, 1)
        self.layout.addWidget(self.button_box, 3, 0, 2, 1)
        self.setLayout(self.layout)

    def save_settings(self) -> None:
        self.temp_config.row_height = self.row_height_box.value()
        self.temp_config.page_len = self.page_len_box.value()
        self.temp_config.sorting_type = SortingType(
            self.sorting_type_box.currentIndex()
        )
        
        self.accept()



from PyQt5.QtWidgets import (
    QMenuBar, QTableWidget, QFrame, QGridLayout, 
    QHeaderView, QLabel, QPlainTextEdit, QAction, 
    QAbstractItemView
)
from PyQt5.QtCore import Qt, QSettings
from PyQt5.QtGui import QIcon, QKeySequence

from plover.engine import StenoEngine
from plover.gui_qt.tool import Tool
from plover.gui_qt.utils import ToolBar

from plover_next_stroke.resources_rc import *
from plover_next_stroke.next_stroke_config import NextStrokeConfig
from plover_next_stroke.config_ui import ConfigUI
from plover_next_stroke.sorting import SortingType


class NextStrokeUI(Tool):
    TITLE = "Next Stroke Suggestions"
    ICON = ":/next_stroke/timeline.svg"
    ROLE = "next_stroke"

    def __init__(self, engine: StenoEngine) -> None:
        super().__init__(engine)
        self.engine: StenoEngine = engine
        self.config = NextStrokeConfig()
        self.restore_state()
        self.show_window()
        self.finished.connect(self.save_state)

    def _restore_state(self, settings: QSettings) -> None:
        if settings.contains("row_height"):
            self.config.row_height = settings.value("row_height", type=int)
        
        self.prev_pin = False
        if settings.contains("pinned") and settings.value("pinned", type=bool):
            self.prev_pin = True
            self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        
        if settings.contains("page_len"):
            self.config.page_len = settings.value("page_len", type=int)
        
        if settings.contains("sorting_type"):
            self.config.sorting_type = SortingType(settings.value("sorting_type", type=int))
        
        if not settings.contains("geometry"):
            self.resize(260, 400)
        
    def _save_state(self, settings: QSettings) -> None:
        settings.setValue("row_height", self.config.row_height)
        settings.setValue("pinned", self.pin_action.isChecked())
        settings.setValue("page_len", self.config.page_len)
        settings.setValue("sorting_type", self.config.sorting_type.value)

    def show_window(self) -> None:
        self.current_label = QLabel(self)
        self.current_label.setText("Current Translation")

        self.current_translation = QPlainTextEdit(self)
        self.current_translation.setFixedHeight(30)
        self.current_translation.setLineWrapMode(True)
        self.current_translation.setReadOnly(True)
        self.current_translation.setPlainText("Awaiting Input")

        self.suggestions_label = QLabel(self)
        self.suggestions_label.setText("Suggestions")

        self.suggestions_table = QTableWidget(self)
        self.suggestions_table.setRowCount(self.config.page_len)
        self.suggestions_table.setColumnCount(2)
        self.suggestions_table.verticalHeader().setDefaultSectionSize(self.config.row_height)
        self.suggestions_table.setMinimumHeight(self.config.row_height * self.config.page_len + self.config.row_height)
        self.suggestions_table.setAlternatingRowColors(True)
        self.suggestions_table.horizontalHeader().hide()
        self.suggestions_table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.suggestions_table.verticalHeader().hide()
        self.suggestions_table.setShowGrid(False)
        self.suggestions_table.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.pin_action = QAction(self)
        self.pin_action.setCheckable(True)
        self.pin_action.setChecked(self.prev_pin)
        self.pin_action.setText("Pin window")
        self.pin_action.setToolTip("Keep this suggestion window on top.")
        self.pin_action.setIcon(QIcon(":/next_stroke/pin.svg"))
        self.pin_action.triggered.connect(self.on_toggle_pin)
        self.pin_action.setShortcut(QKeySequence("Ctrl+P"))

        self.settings_action = QAction(self)
        self.settings_action.setText("Next Stroke Suggestions settings")
        self.settings_action.setText("Configure Next Stroke Suggestions.")
        self.settings_action.setIcon(QIcon(":/next_stroke/settings.svg"))
        self.settings_action.triggered.connect(self.on_settings)
        self.settings_action.setShortcut(QKeySequence("Ctrl+S"))

        self.page_label = QLabel(self)
        self.page_label.setText("Page 0 of 0")
        self.page_label.setAlignment(Qt.AlignHCenter)

        self.layout = QGridLayout()
        self.layout.addWidget(self.current_label, 0, 0, 1, 2)
        self.layout.addWidget(self.current_translation, 1, 0, 1, 2)
        self.layout.addWidget(self.suggestions_label, 2, 0, 1, 2)
        self.layout.addWidget(self.suggestions_table, 3, 0, 1, 2)
        self.layout.addWidget(ToolBar(
            self.pin_action,
            self.settings_action
        ), 4, 0)
        self.layout.addWidget(self.page_label, 4, 1)
        self.setLayout(self.layout)

        self.show()
    
    def on_toggle_pin(self, _: bool = False) -> None:
        flags = self.windowFlags()

        if self.pin_action.isChecked():
            flags |= Qt.WindowStaysOnTopHint
        else:
            flags &= ~Qt.WindowStaysOnTopHint

        self.setWindowFlags(flags)
        self.show()

    def on_settings(self, *args) -> None:
        config_dialog = ConfigUI(self.config.copy(), self)
        if config_dialog.exec():
            self.config = config_dialog.temp_config
            self.suggestions_table.setRowCount(self.config.page_len)
            self.suggestions_table.verticalHeader().setDefaultSectionSize(self.config.row_height)
            self.suggestions_table.setMinimumHeight(self.config.row_height * self.config.page_len + self.config.row_height)

    def get_next_stroke_config(self) -> NextStrokeConfig:
        return self.config



from plover_next_stroke.sorting import SortingType


CONFIG_ITEMS = {
    "row_height": 30,
    "page_len": 10,
    "sorting_type": SortingType.FREQUENCY
}


class NextStrokeConfig:
    def __init__(self, values: dict = None):
        if values is None:
            values = dict()

        for key, default in CONFIG_ITEMS.items():
            if key in values:
                setattr(self, key, values[key])
            else:
                setattr(self, key, default)

    def copy(self) -> "NextStrokeConfig":
        value_dict = {k: getattr(self, k) for k in CONFIG_ITEMS.keys()}
        return NextStrokeConfig(value_dict)



from plover.translation import Translator
from plover.steno import Stroke


def prev_page(translator: Translator, stroke: Stroke, argument: str):
    translator.next_stroke_state = "prev_page"


def next_page(translator: Translator, stroke: Stroke, argument: str):
    translator.next_stroke_state = "next_page"


def next_stroke_reload(translator: Translator, stroke: Stroke, argument: str):
    translator.next_stroke_state = "next_stroke_reload"



from PyQt5.QtWidgets import QTableWidgetItem

from typing import Dict, Tuple, List, Optional, Any

from plover import system
from plover.engine import StenoEngine
from plover.registry import registry
from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.translation import Translation

from plover_next_stroke.next_stroke_ui import NextStrokeUI
from plover_next_stroke.sorting import SortingType, get_sorter, sort_suggestions


STROKE_TYPE = str
OUTLINE_TYPE = Tuple[STROKE_TYPE]


class TranslationNode:
    def __init__(self, translation: str = "") -> None:
        self.translation = translation
        self.children: Dict[STROKE_TYPE, "TranslationNode"] = {}

    def add_child(self, outline: OUTLINE_TYPE, translation: str) -> None:
        if not outline:
            return
        
        outline_len = len(outline)
        outline_head = outline[0]

        if outline_len == 1:
            if outline_head in self.children:
                self.children[outline_head].translation = translation
            else:
                self.children[outline_head] = TranslationNode(translation)
        else:
            outline_tail = outline[1:]
            if outline_head not in self.children:
                self.children[outline_head] = TranslationNode()
            
            self.children[outline_head].add_child(outline_tail, translation)
    
    def get_node(self, outline: OUTLINE_TYPE) -> Optional["TranslationNode"]:
        if not outline:
            return self

        outline_head = outline[0]
        if outline_head not in self.children:
            return None
        else:
            return self.children[outline_head].get_node(outline[1:])

    def get_suggestions(self) -> List[Tuple[OUTLINE_TYPE, str]]:
        suggestions_list = []

        if self.children:
            for stroke, node in self.children.items():
                if node.translation:
                    suggestions_list.append(([stroke], node.translation))

                node_suggestions = node.get_suggestions()
                for outline, translation in node_suggestions:
                    suggestions_list.append(([stroke] + outline, translation))
        
        return suggestions_list


class NextStrokeSuggestions(NextStrokeUI):
    def __init__(self, engine: StenoEngine) -> None:
        super().__init__(engine)

        self._translate_tree = None
        self._suggestions: List[Tuple[OUTLINE_TYPE, str]] = []
        self._prev_node: Optional[TranslationNode] = None
        self._page = 0

        self._stroke_formatter: Optional[Callable[[STROKE_TYPE], STROKE_TYPE]] = None
        self._translation_formatter: Optional[Callable[[str], str]] = None
        self._system_sorter: Optional[Callable[[Tuple[OUTLINE_TYPE, str]], Any]] = None

        engine.signal_connect("stroked", self.on_stroke)
        engine.signal_connect("dictionaries_loaded", self.on_dict_update)
        engine.signal_connect("config_changed", self.on_config_changed)
        engine.signal_connect("add_translation", self.on_dict_update)
        self.index_dictionaries()
        self.on_config_changed()

    def update_table(self) -> None:
        top_index = self._page * self.config.page_len

        page_count = (len(self._suggestions) - 1) // self.config.page_len + 1
        displayed = self._suggestions[top_index:top_index + self.config.page_len]
        display_len = len(displayed)

        for index, (outline, translation) in enumerate(displayed):
            self.suggestions_table.setItem(index, 0, QTableWidgetItem("/".join(outline)))
            self.suggestions_table.setItem(index, 1, QTableWidgetItem(translation))
        
        if display_len < self.config.page_len:
            for index in range(display_len, self.config.page_len):
                self.suggestions_table.setItem(index, 0, QTableWidgetItem(""))
                self.suggestions_table.setItem(index, 1, QTableWidgetItem(""))
        
        self.page_label.setText(f"Page {self._page + 1} of {page_count}")

    def on_stroke(self, _: tuple) -> None:
        update_suggestions = True

        if hasattr(self.engine._translator, "next_stroke_state"):
            next_stroke_state = self.engine._translator.next_stroke_state
            max_pages = (len(self._suggestions) - 1) // self.config.page_len + 1

            if next_stroke_state == "prev_page":
                self._page = (self._page - 1) % max_pages
                update_suggestions = False
            
            elif next_stroke_state == "next_page":
                self._page = (self._page + 1) % max_pages
                update_suggestions = False
            
            elif next_stroke_state == "next_stroke_reload":
                self.index_dictionaries()
            
            self.engine._translator.next_stroke_state = ""

        prev_translations: List[Translation] = self.engine.translator_state.prev()

        if prev_translations is not None and prev_translations and update_suggestions:
            translation = prev_translations[-1]
            current_outline = translation.rtfcre
            current_output = translation.english

            if current_output is None:
                self.current_translation.setPlainText(
                    f"{'/'.join(current_outline)}"
                )
            else:
                self.current_translation.setPlainText(
                    f"{'/'.join(current_outline)}: {current_output}"
                )

            tree_node = self._translate_tree.get_node(current_outline)
            queued_suggestions = []

            if tree_node is not None:
                suggestions = tree_node.get_suggestions()
                
                if self._prev_node is not None and current_outline:
                    current_stroke = current_outline[-1]
                    traced_node = self._prev_node.get_node([current_stroke])
                    if traced_node is not None and traced_node is not tree_node:
                        suggestions += traced_node.get_suggestions()

                queued_suggestions = sort_suggestions(
                    suggestions=suggestions,
                    sorting_type=self.config.sorting_type,
                    stroke_formatter=self._stroke_formatter,
                    translation_formatter=self._translation_formatter,
                    system_sorter=self._system_sorter
                )

                self._prev_node = tree_node

            self._suggestions = queued_suggestions
            self._page = 0
        
        self.update_table()
    
    def index_dictionaries(self) -> None:
        self._translate_tree = TranslationNode()
        dictionaries: StenoDictionaryCollection = self.engine.dictionaries

        dictionary: StenoDictionary
        for dictionary in dictionaries.dicts:
            if dictionary.enabled:
                for outline, translation in dictionary.items():
                    self._translate_tree.add_child(outline, translation)

    def on_dict_update(self) -> None: 
        self.index_dictionaries()
    
    def on_config_changed(self) -> None:
        system_name = system.NAME
        system_mod = registry.get_plugin("system", system_name).obj

        self._stroke_formatter = getattr(system_mod, "NS_STROKE_FORMATTER", None)
        self._translation_formatter = getattr(system_mod, "NS_TRANSLATION_FORMATTER", None)
        self._system_sorter = getattr(system_mod, "NS_SORTER", None)



from enum import Enum
from typing import Tuple, Union, Callable, Any, List, Optional

from plover import system


STROKE_TYPE = str
OUTLINE_TYPE = Tuple[STROKE_TYPE]


class SortingType(Enum):
    FREQUENCY = 0
    FREQUENCY_NUM = 1
    FREQUENCY_ALPHA = 2
    STROKE_COUNT = 3
    ALPHABETICAL = 4
    SYSTEM_DEFINED = 5


sorting_descriptions = [
    "Frequency",
    "Frequency (Prioritize Numbers)",
    "Frequency (Prioritize Non-numeric)",
    "Stroke Count",
    "Alphabetical",
    "System Defined"
]


def to_int(string: str, default: int) -> int:
    try:
        return int(string)
    except ValueError:
        return default


def num_score(outline: OUTLINE_TYPE) -> Tuple[int, ...]:
    return tuple(to_int(s, 999999) for s in outline)


def get_sorter(sorting_type: SortingType) -> Callable[[Tuple[OUTLINE_TYPE, str]], Any]:
    if sorting_type == SortingType.FREQUENCY:
        if system.ORTHOGRAPHY_WORDS is not None:
            return lambda s: (len(s[0]), system.ORTHOGRAPHY_WORDS.get(s[1], 999999))
        else:
            return lambda s: len(s[0])

    elif sorting_type == SortingType.FREQUENCY_NUM:
        if system.ORTHOGRAPHY_WORDS is not None:
            return lambda s: (num_score(s[0]), system.ORTHOGRAPHY_WORDS.get(s[1], 999999))
        else:
            return lambda s: num_score(s[0])

    elif sorting_type == SortingType.FREQUENCY_ALPHA:
        if system.ORTHOGRAPHY_WORDS is not None:
            return lambda s: (not s[0][-1].isalpha(), len(s[0]), system.ORTHOGRAPHY_WORDS.get(s[1], 999999))
        else:
            return lambda s: (not s[0][-1].isalpha(), len(s[0]))

    elif sorting_type == SortingType.STROKE_COUNT:
        return lambda s: len(s[0])

    elif sorting_type == SortingType.ALPHABETICAL:
        return lambda s: s[1].lower()


def sort_suggestions(
    suggestions: List[Tuple[OUTLINE_TYPE, str]], 
    sorting_type: SortingType,
    stroke_formatter: Optional[Callable[[STROKE_TYPE], STROKE_TYPE]] = None,
    translation_formatter: Optional[Callable[[str], str]] = None,
    system_sorter: Optional[Callable[[Tuple[OUTLINE_TYPE, str]], Any]] = None
) -> List[Tuple[OUTLINE_TYPE, str]]:
    result = []
    for outline, translation in suggestions:
        if stroke_formatter is not None:
            outline = tuple(stroke_formatter(s) for s in outline)
        if translation_formatter is not None:
            translation = translation_formatter(translation)
        
        result.append((outline, translation))
    
    if sorting_type == SortingType.SYSTEM_DEFINED:
        if system_sorter is not None:
            return sorted(result, key=system_sorter)
        
        sorting_type = SortingType.FREQUENCY
    
    return sorted(result, key=get_sorter(sorting_type))
